require("verlet")
require("rocket")
game = {}
game.ojects = {}
game.objects = {}
game._objectsToSpawn = {}
game._objectsToDie = {}
game._cnt = 1

function game.reset()

    game.ojects = {}
    game.objects = {}
    game._objectsToSpawn = {}
    game._objectsToDie = {}
    game._cnt = 1
end

function game.createRocket(x, y, fuel)
    local rocket = Rocket:new(game._cnt, x, y, fuel)
    table.insert(game._objectsToSpawn, rocket)
    game._cnt = game._cnt + 1

    return rocket

end

function game.createExplosion(x, y)
    local explosion = {id = game._cnt, x=x, y=y}
    table.insert(game._objectsToSpawn, explosion)
    game._cnt = game._cnt + 1

    explosion.smoke = love.graphics.newParticleSystem(All_Images.explosion, 32)
    explosion.smoke:setParticleLifetime(2, 4) 
    explosion.smoke:setSpeed(20,50)
    explosion.smoke:setSpread(math.pi * 2)    
    explosion.smoke:setRotation(-math.pi,math.pi)    
    explosion.smoke:setColors(1, 1, 1, 1, 1, 1, 1, 0) 
    
    explosion.smoke:emit(10)

    explosion.flash = love.graphics.newParticleSystem(All_Images.explosionFlash, 16)
    explosion.flash:setParticleLifetime(0.1,0.3) 
    explosion.flash:setSpeed(20,70)
    explosion.flash:setSpread(math.pi * 2)    
    explosion.flash:emit(10)
    explosion.flash:setRotation(-math.pi,math.pi)
    explosion.flash:setColors(1, 1, 1, 1, 1, 1, 1, 0) 


      function explosion.render(self)
        love.graphics.setColor(1,1,1,1)
        love.graphics.draw(self.flash, self.x, self.y)
        love.graphics.draw(self.smoke, self.x, self.y)
    end

    function explosion.update(self, dt)
        self.flash:update(dt)
        self.smoke:update(dt)
    end
    return explosion
end

function game.createArea(x, y, radius, callback)
    local area = {
        id = game._cnt,
        x = x,
        y = y,
        radius = radius,
        callback = callback
    }
    table.insert(game._objectsToSpawn, area)
    game._cnt = game._cnt + 1

    function area.render(self)
        love.graphics.setColor(0, 1, 0, 0.5)
        love.graphics.circle("fill", self.x, self.y, self.radius)
    end

    function area.update(self, dt)
        for i, v in ipairs(game.objects) do
            if v ~= self and v.physicsObj then
                if v.physicsObj:center():distance(Vector:new(self.x, self.y)) <
                    self.radius and callback then
                    callback(self, v)
                end
            end
        end
    end

    return area;
end

function game.createLevelFromTileData(offsetX, offsetY, tiles, border)

    local width = 150
    local height = 150

    game.decorationCanvas = love.graphics.newCanvas(
                                width * (#tiles + 1) + offsetX,
                                height * (#tiles[1] + 1) + offsetY)
    love.graphics.setCanvas(game.decorationCanvas)
    love.graphics.clear(0, 0, 0, 0)
    love.graphics.setColor(1, 1, 1, 1)

    for x = 1, #tiles do
        for y = 1, #tiles[x] do

            if tiles[y][x] >= 16 then
                local verts = sat.verts0((x) * width + offsetX,
                                         (y) * height + offsetY, width, height)

                love.graphics.draw(All_Images.caveBackground, verts[1].x,
                                   verts[1].y)
            elseif tiles[y][x] >= 0 and tiles[y][x] <= 15 then
                local o = {id = game._cnt}
                game._cnt = game._cnt + 1
                local physicsObj = verlet.newRect((x) * width + offsetX,
                                                  (y) * height + offsetY, width,
                                                  height, 0)
                verlet.pin(physicsObj, nil)
                o.physicsObj = physicsObj

                o.type = tiles[y][x]

                love.graphics.draw(All_Images.rock.spritesheet,
                                   All_Images.rock.quads[o.type],
                                   physicsObj.verts[1].x, physicsObj.verts[1].y)

                table.insert(game._objectsToSpawn, o)
           end

        end
    end

    love.graphics.setCanvas()

end

function game.createFuelCanister(x, y)
    local callback = function(area, obj)
        if obj == player then 
            player:refuel(area.fuelAmount) 
            game.removeObj(area)
        end
    end
    local fuelCanister = game.createArea(x, y, 30, callback)
    fuelCanister.fuelAmount = 100
    fuelCanister.randomRotationOffset = math.random(360)
    fuelCanister.render = function(self)
        love.graphics.draw(All_Images.fuelCanister, self.x, self.y, (love.timer.getTime()+ self.randomRotationOffset) * 2 * math.pi / 5,1,1,15,23)
    end
end

function game.generateItems(offsetX, offsetY, level, maxItemCount, tilesize)
    local possibleItemSpawns = {}
    local tiles = level.level
    for x = 1, #tiles do
        for y = 1, #tiles[x] do
            if tiles[y][x] == 17 and not (x == level.furthestPoint.y and y == level.furthestPoint.x) then
                -- adding at random position so we can simply pop the last element x times later
                table.insert(possibleItemSpawns,
                             math.random(1, math.max(1,#possibleItemSpawns)), {x = x, y = y})
            end
        end
    end

    for i = 1, maxItemCount do
        if #possibleItemSpawns == 0 then break end
        game.createFuelCanister(possibleItemSpawns[#possibleItemSpawns].x *
                                    tilesize + offsetX  + 75,
                                possibleItemSpawns[#possibleItemSpawns].y *
                                    tilesize + offsetY + 75)
        table.remove(possibleItemSpawns)
    end
end

function game.renderGeometryObj(obj, l, t, w, h)
    local verts = obj.physicsObj.verts

    if l - verts[1].x < w + obj.physicsObj.colissioncheckradius and t -
        verts[1].y < h + obj.physicsObj.colissioncheckradius then

        love.graphics.draw(All_Images.rock.spritesheet,
                           All_Images.rock.quads[obj.type], verts[1].x,
                           verts[1].y)

    end

end

function game.createStalagtite(x1, y1, x2, y2, x3, y3)
    local o = {id = game._cnt}
    o.physicsObj = verlet.newTriangle(x1, y1, x2, y2, x3, y3)
    o.color = {r = 0.43, g = 0.45, b = 0.45, a = 1}
    o.render = function(self)
        local verts = self.physicsObj.verts
        love.graphics.setColor(self.color.r, self.color.g, self.color.b,
                               self.color.a)
        love.graphics.polygon("fill", verts[1].x, verts[1].y, verts[2].x,
                              verts[2].y, verts[3].x, verts[3].y)
    end
    verlet.pin(o.physicsObj, nil)
    table.insert(game._objectsToSpawn, o)
    return o

end

function game.removeObj(obj)
    table.insert(game._objectsToDie, obj)
    if obj.physicsObj then verlet.removeObj(obj.physicsObj) end
end

function game._checkObjectsToSpawn()
    for i, v in ipairs(game._objectsToSpawn) do table.insert(game.objects, v) end
    game._objectsToSpawn = {}
end

function game._remove(obj)
    for i = 1, #game.objects do
        if game.objects[i].id == obj.id then
            table.remove(game.objects, i)
            return
        end
    end
end

function game._checkObjectsToDie()
    for i, v in ipairs(game._objectsToDie) do game._remove(v) end
    game._objectsToDie = {}
end

function game.update(dt)

    game._checkObjectsToSpawn()
    game._checkObjectsToDie()
    verlet.game_update(dt)
    for i, v in ipairs(game.objects) do if v.update then v:update(dt) end end

end

function game.render(l, t, w, h)

    if game.background then love.graphics.draw(game.background, 0, 0) end
    if game.decorationCanvas then
        love.graphics.draw(game.decorationCanvas, 0, 0)
    end
    for i, o in ipairs(game.objects) do
        if o.render then o:render(l, t, w, h) end
    end
end
