require("vector")
require("verlet")
input = {}
-- ammount a controler stick must at least be moved until we recognize it as an input
input.joystickdeadzone = 0.1

input.state = "DEFAULT"

input.keymap = {
    up = "up",
    down = "down",
    left = "left",
    right = "right",
    space = "continue",
    escape = "exit",
    kpenter = "start",
    backspace = "delete"
}
input.keymap["return"] = "start"
input.controllermap = {
    a = "continue",
    start = "start",
    dpup = "up",
    dpdown = "down",
    dpleft = "left",
    dpright = "right"
}

function love.mousepressed(x, y, button, istouch, presses) 
        if gamestate.mousepressed then gamestate.mousepressed(x, y, button) end
end

function love.mousemoved(x, y, dx, dy)
    if gamestate.mousemoved then gamestate.mousemoved(x, y, dx, dy) end
end

function love.mousereleased(x, y, button, istouch, presses) end

function love.wheelmoved(x, y)
    if gamestate.wheelmoved then gamestate.wheelmoved(x, y) end
end

function love.keypressed(key)

    local mapping = input.keymap[key]
    if mapping == "exit" and not gamestate.handlequit then love.event.quit() end

    if gamestate.oninput then
        -- Let the state handle this

        gamestate.oninput(mapping or key)
        return
    end

end

function love.gamepadpressed(joystick, button)
    if gamestate.oninput then
        -- Let the state handle this
        local mapping = input.controllermap[button]

        gamestate.oninput(mapping or key)

    end
end

function input.update(dt)
    if not player then return end

    local moving = false
    if love.keyboard.isDown("up") then
        player:speedup(dt, 1)
        moving = true
    end
    if love.keyboard.isDown("right") then
        player:turnright(dt, 1)
        moving = true
    end
    if love.keyboard.isDown("left") then
        player:turnleft(dt, 1)
        moving = true
    end
    if love.keyboard.isDown("down") then
        player:reverse(dt, 1)
        moving = true
    end

    if p1joystick ~= nil then moving = moving or input.handleJoystick(dt) end

    if not moving then player:notmoving() end

end

function input.handleJoystick(dt)
    local movementx = p1joystick:getGamepadAxis("leftx")
    local rtriggeramount = p1joystick:getGamepadAxis("triggerright")
    local ltriggeramount = p1joystick:getGamepadAxis("triggerleft")

    local moving = false

    if movementx < -input.joystickdeadzone then
        moving = true
        player:turnleft(dt, -movementx)
    end
    if movementx > input.joystickdeadzone then
        moving = true
        player:turnright(dt, movementx)
    end

    if math.abs(rtriggeramount) > input.joystickdeadzone then
        moving = true
        player:speedup(dt, rtriggeramount)
    end

    if math.abs(rtriggeramount) > input.joystickdeadzone then
        moving = true
        player:reverse(dt, rtriggeramount)
    end

    return moving

end
