local levelgen = {}
function levelgen.initGrid(xsize, ysize)
    local level = {}
    for x = 1, xsize do
        level[x] = {}
        for y = 1, ysize do level[x][y] = "#" end
    end
    return level
end

function levelgen.copy(level)
    local newlevel = {}
    for x = 1, #level do
        newlevel[x] = {}
        for y = 1, #level[x] do newlevel[x][y] = level[x][y] end
    end
    return newlevel
end

function levelgen.initGridDistances(xsize, ysize)
    local level = {}
    for x = 1, xsize do
        level[x] = {}
        for y = 1, ysize do level[x][y] = math.huge end
    end
    return level
end

levelgen.drunkard = {}

levelgen.drunkard.steps = 50

function levelgen.drunkard.istGueltigUndWand(level, x, y)
    return x >= 2 and y >= 2 and x <= #level - 2 and y <= #level[x] - 2 and
               level[x][y] == "#"
end

function levelgen.drunkard.istGueltig(level, x, y)
    return x >= 2 and y >= 2 and x <= #level - 2 and y <= #level[x] - 2
end

function levelgen.drunkard.gueltigeRichtungen(level, x, y, lastXdir, lastYDir)
    local gueltigeRichtungen = {}
    if levelgen.drunkard.istGueltigUndWand(level, x - 1, y) then
        table.insert(gueltigeRichtungen, {x = x - 1, y = y})
    end
    if levelgen.drunkard.istGueltigUndWand(level, x + 1, y) then
        table.insert(gueltigeRichtungen, {x = x + 1, y = y})
    end
    if levelgen.drunkard.istGueltigUndWand(level, x, y - 1) then
        table.insert(gueltigeRichtungen, {x = x, y = y - 1})
    end
    if levelgen.drunkard.istGueltigUndWand(level, x, y + 1) then
        table.insert(gueltigeRichtungen, {x = x, y = y + 1})
        table.insert(gueltigeRichtungen, {x = x, y = y + 1})
    end
    if #gueltigeRichtungen == 0 then
        if levelgen.drunkard.istGueltig(level, x - 1, y) then
            table.insert(gueltigeRichtungen, {x = x - 1, y = y})
        end
        if levelgen.drunkard.istGueltig(level, x + 1, y) then
            table.insert(gueltigeRichtungen, {x = x + 1, y = y})
        end
        if levelgen.drunkard.istGueltig(level, x, y - 1) then
            table.insert(gueltigeRichtungen, {x = x, y = y - 1})
        end
        if levelgen.drunkard.istGueltig(level, x, y + 1) then
            table.insert(gueltigeRichtungen, {x = x, y = y + 1})
        end
    end
    return gueltigeRichtungen
end

function levelgen.drunkard._isSurroundedByWalls(level, x, y)
    for i = -1, 1 do
        for j = -1, 1 do
            if x + i < 1 or y + j < 1 or x + i > #level or y + j > #level[x] or
                level[x + i][y + j] == " " then return false end
        end
    end

    return true
end

function levelgen.drunkard.pass2(result)
    local level = result.level
    for x = 1, #level do
        for y = 1, #level[x] do
            if levelgen.drunkard._isSurroundedByWalls(level, x, y) then
                level[x][y] = "+"
            end

        end
    end
    return result
end

function levelgen.calcDistance(posX, posY, distances)
    local lowestDistance = math.huge

    for x = -1, 1 do
        for y = -1, 1 do
            if distances[posX + x][posY + y] < lowestDistance then
                lowestDistance = distances[posX + x][posY + y]
            end

        end
    end
    if lowestDistance == math.huge then
        return 0
    else
        return lowestDistance + 1
    end

end

-- Using the strategy discribed here: http://pcg.wikidot.com/pcg-algorithm:drunkard-walk. But i actually implemented a bug!
function levelgen.drunkard.generate(size, maxSteps, startx, starty, stepstaken)
    local level = levelgen.initGrid(size, size)
    local steps = maxSteps or levelgen.drunkard.steps
    local distances = levelgen.initGridDistances(size, size)

    startx = startx or 1
    starty = starty or 1
    level[startx][starty] = " "
    local x = startx
    local y = starty

    local letzteRichtung = {x = 0, y = 0}
    local furthestPointSoFar = {x, y}
    local furthestDistance = 0
    for i = 1, steps do

        local richtungen = levelgen.drunkard.gueltigeRichtungen(level, x, y)
        local bias = nil
        for i, v in ipairs(richtungen) do
            if v.x - x == letzteRichtung.x and v.y - y == letzteRichtung.y then
                bias = v
                break
            end
        end
        if bias then table.insert(richtungen, bias) end
        local nextCell = richtungen[math.random(#richtungen)]
        letzteRichtung.x = nextCell.x - x
        letzteRichtung.y = nextCell.y - y
        x = nextCell.x
        y = nextCell.y

        distances[x][y] = levelgen.calcDistance(x, y, distances)

        if distances[x][y] > furthestDistance then

            furthestDistance = distances[x][y]
            furthestPointSoFar = {x = x, y = y}
        end
        level[x][y] = " "

    end

    return levelgen.drunkard.pass2({
        level = level,
        furthestPoint = furthestPointSoFar
    })

end

levelgen.cellular = {}
levelgen.cellular.settings = {}
levelgen.cellular.settings.initPosibillity = 50
levelgen.cellular.settings.stayaliveTreshold = 2
levelgen.cellular.settings.bornCount = 3
levelgen.cellular.settings.overCrowdingCount = 5
-- Harder version: 
-- levelgen.cellular.settings.stayaliveTreshold = 2
-- levelgen.cellular.settings.bornCount = 5
-- levelgen.cellular.settings.overCrowdingCount = 6

levelgen.cellular.settings.minCaveSize = 6

function levelgen.cellular.neighborcount(level, x, y)
    local neighborcount = 0
    for i = -1, 1 do
        for j = -1, 1 do
            if x + i > 1 and y + j > 1 and x + i < #level and y + j < #level[x] and
                level[x + i][y + j] == "#" then
                neighborcount = neighborcount + 1
            end
        end
    end
    return neighborcount
end

function levelgen.floodfill(level, visited, startx, starty)
    local cells = {}
    if startx < 1 or starty < 1 or startx > #level or starty > #level[startx] then
        return cells
    end

    if visited[startx][starty] then return cells end

    visited[startx][starty] = true
    if level[startx][starty] == " " then

        table.insert(cells, {x = startx, y = starty})
        utils.addAll(cells,
                     levelgen.floodfill(level, visited, startx - 1, starty))
        utils.addAll(cells,
                     levelgen.floodfill(level, visited, startx + 1, starty))
        utils.addAll(cells,
                     levelgen.floodfill(level, visited, startx, starty - 1))
        utils.addAll(cells,
                     levelgen.floodfill(level, visited, startx, starty + 1))
    end
    return cells
end

function levelgen.calcDistances(level, startx, starty)
    local distances = {}
    for x = 1, #level do
        distances[x] = {}
        for y = 1, #level[x] do distances[x][y] = math.huge end
    end
    distances[startx][starty] = 0

    local distancesToCalculate = {{x = startx, y = starty}}

    while (#distancesToCalculate > 0) do
        local currentPos = distancesToCalculate[1]
        table.remove(distancesToCalculate, 1)
        local neighbors = {
            {x = currentPos.x - 1, y = currentPos.y},
            {x = currentPos.x + 1, y = currentPos.y},
            {x = currentPos.x, y = currentPos.y - 1},
            {x = currentPos.x, y = currentPos.y + 1}
        }
        for i, v in ipairs(neighbors) do
            local x = v.x
            local y = v.y

            if levelgen.isPassable(level, x, y) and distances[x][y] >
                distances[currentPos.x][currentPos.y] + 1 then
                table.insert(distancesToCalculate, {x = x, y = y})
                distances[x][y] = distances[currentPos.x][currentPos.y] + 1
            end
        end

    end
    return distances

end

function levelgen.findFurthestPoint(level, startx, starty)
    local furthestDistance = 0
    local furthestPointSoFar
    local distances = levelgen.calcDistances(level, startx, starty)
    for x = 1, #distances do
        for y = 1, #distances[x] do
            if distances[x][y] ~= math.huge and distances[x][y] >
                furthestDistance then
                furthestDistance = distances[x][y]
                furthestPointSoFar = {x = x, y = y}
            end
        end
    end
    return furthestPointSoFar
end

function levelgen.cellular.pass2(result, startx, starty)

    local level = result.level
    local visited = {}
    for x = 1, #level do
        visited[x] = {}
        for y = 1, #level[x] do visited[x][y] = false end
    end
    local caves = {}
    for x = 1, #level do
        for y = 1, #level[x] do
            if not visited[x][y] then

                local cells = levelgen.floodfill(level, visited, x, y)

                -- fill if cave is too small
                if #cells < levelgen.cellular.settings.minCaveSize then

                    for i, pos in ipairs(cells) do

                        level[pos.x][pos.y] = "#"
                    end
                else
                    table.insert(caves, cells)
                end
            end
        end
    end
    if level[startx][starty] ~= " " then
        level[startx][starty] = " "
        table.insert(caves, 1, {{x = startx, y = starty}})
    end

    levelgen.cellular.connectCaves(level, caves)

    result.furthestPoint = levelgen.findFurthestPoint(level, startx, starty)

    for x = 1, #level do
        for y = 1, #level[x] do
            if levelgen.drunkard._isSurroundedByWalls(level, x, y) then
                level[x][y] = "+"
            end

        end
    end

    return result
end

function levelgen.cellular.connectCaves(level, caves)

    for i = 1, #caves do
        -- debugging
        if caves[i + 1] ~= nil then

            levelgen.cellular.connect2Caves(level, caves[i], caves[i + 1])
        end

    end
end

function levelgen.cellular.connect2Caves(level, cave1, cave2)
    local closestC1 = nil
    local closestC2 = nil
    local closestDist = math.huge
    for i = 1, #cave1 do
        for j = 1, #cave2 do
            local distSquared = (cave1[i].x - cave2[j].x) ^ 2 +
                                    (cave1[i].y - cave2[j].y) ^ 2
            if distSquared < closestDist then
                closestC1 = cave1[i]
                closestC2 = cave2[j]
                closestDist = distSquared
            end
        end
    end
    local currentX = closestC1.x
    local currentY = closestC1.y

    local dirx = 1
    if closestC1.x > closestC2.x then dirx = -1 end
    for x = closestC1.x, closestC2.x, dirx do
        currentX = x

        level[currentX][currentY] = " "

    end
    local diry = 1
    if closestC1.y > closestC2.y then diry = -1 end
    for y = closestC1.y, closestC2.y, diry do
        currentY = y

        level[currentX][currentY] = " "
    end
end

function levelgen.cellular.getTileType(level, x, y)
    local mask_left = 1
    local mask_bottom = 2
    local mask_right = 4
    local mask_top = 8

    if (level[x][y]) == " " then return 16 end

    local currentTileMask = 15

    if not levelgen.cellular.isFree(level, x - 1, y) then
        currentTileMask = utils.bxor(currentTileMask, mask_top)
    end
    if not levelgen.cellular.isFree(level, x + 1, y) then
        currentTileMask = utils.bxor(currentTileMask, mask_bottom)
    end

    if not levelgen.cellular.isFree(level, x, y + 1) then
        currentTileMask = utils.bxor(currentTileMask, mask_right)
    end

    if not levelgen.cellular.isFree(level, x, y - 1) then
        currentTileMask = utils.bxor(currentTileMask, mask_left)
    end

    return currentTileMask
end

function levelgen.inLevel(level, x, y)
    return x > 0 and y > 0 and x <= #level and y <= #level[x]
end

function levelgen.cellular.calculateItemSpawns(result)
    local function reachableNeighborWalls(level, posX, posY)
        local neighbors = {
            {x = posX - 1, y = posY}, {x = posX + 1, y = posY},
            {x = posX, y = posY - 1}, {x = posX, y = posY + 1}
        }
        local neighborcount = 0
        for i, v in ipairs(neighbors) do
            if levelgen.inLevel(level, v.x, v.y) and level[v.x][v.y] < 15 then
                neighborcount = neighborcount + 1
            end
        end
        return neighborcount

    end
    local level = result.level
    for x = 1, #level do
        for y = 1, #level[x] do
            if level[x][y] == 16 and
                reachableNeighborWalls(level, x, y) == 3 then
                level[x][y] = 17
            end
        end
    end
end

function levelgen.cellular.decorate(result)
    local level = result.level
    local decoratedLevel = {}
    for x = 1, #level do
        decoratedLevel[x] = {}
        for y = 1, #level[x] do
            decoratedLevel[x][y] = levelgen.cellular.getTileType(level, x, y)
        end
    end
    result.level = decoratedLevel

    levelgen.cellular.calculateItemSpawns(result)
    return result
end

function levelgen.cellular.isFree(level, x, y)
    return x < 1 or y < 1 or x > #level or y > #level[x] or level[x][y] == " "
end

function levelgen.isPassable(level, x, y)
    return x > 0 and y > 0 and x <= #level and y <= #level[x] and level[x][y] ==
               " "
end

function levelgen.cellular.generate(size, generations, startx, starty)
    local level = levelgen.initGrid(size, size)
    local xsize = size
    local ysize = size
    startx = startx or 1
    starty = starty or 1
    for x = 2, xsize - 1 do
        for y = 2, ysize - 1 do
            if math.random(100) <= levelgen.cellular.settings.initPosibillity then
                level[x][y] = " "
            end
        end
    end
    for i = 1, generations do

        local newlevel = levelgen.copy(level)
        for x = 2, xsize - 1 do
            for y = 2, ysize - 1 do

                local neighborcount = levelgen.cellular.neighborcount(level, x,
                                                                      y)
                if level[x][y] == " " then
                    if neighborcount >= levelgen.cellular.settings.bornCount then
                        newlevel[x][y] = "#"
                    end
                else
                    if neighborcount <
                        levelgen.cellular.settings.stayaliveTreshold or
                        neighborcount >
                        levelgen.cellular.settings.overCrowdingCount then
                        newlevel[x][y] = " "
                    end
                end

            end
        end
        level = newlevel
    end
    return levelgen.cellular.decorate(levelgen.cellular.pass2({level = level},
                                                              startx, starty))
end

function levelgen.rendercell(cell)
    if cell == 16 then
        return "  "
    elseif cell == 17 then
        return "**"
    elseif cell == "15" then
        return string.format("%02d", cell)
    else
        return string.format("%02d", cell)
    end
end

function levelgen.printlevel(level)

    for x = 1, #level do
        local line = ""
        for y = 1, #level[x] do
            line = line .. levelgen.rendercell(level[x][y])
        end
        print(line)
    end
end

return levelgen
