require("verlet")
require("vector")
require("utils")
require("input")
require("game")
local mainmenu = require("mainmenu")
local leveldone = require("leveldone")
local upgrade = require("upgrade")
local sounsdstuff = require("soundstuff")
local credits = require("credits")
local paused = require("paused")

local rougelike = require("rougelike")
stars = require("stars")

gamera = require("lib/gamera")
levelgen = require("levelgen")

GAMESTATES = {}

GAMESTATES.default = {handlequit = true}
GAMESTATES.default.render = function(self)

    if camera then
        camera:draw(game.render)
    else
        game.render(0, 0, love.graphics.getWidth(), love.graphics.getHeight())
    end

    love.graphics.setColor(1, 1, 1, 1)
    -- renderFuelGauge
    local gaugeWidth = math.max(100, love.graphics.getWidth() / 10)
    love.graphics.setLineWidth(2)
    love.graphics.setColor(0.9, 0.9, 0.9, 1)
    love.graphics.rectangle("line", 20, 20, gaugeWidth, 24)
    love.graphics.setColor(21 / 255, 71 / 255, 52 / 255, 1)
    love.graphics.rectangle("fill", 22, 22,
                            (gaugeWidth - 2) * (player.fuel / player.maxFuel),
                            20)
    love.graphics.setColor(1, 1, 1, 1)
    love.graphics.print("Level: " .. currentRun.level, 0,
                        love.graphics.getHeight() - 30)

    love.graphics.print("FPS: " .. love.timer.getFPS(),
                        love.graphics.getWidth() - 100, 30)
end
function GAMESTATES.default.oninput(key)
    if key == "exit" or key == "start" then
        gamestate = GAMESTATES.paused
        paused.init()
    end
end
GAMESTATES.default.ongameover = function(isWin, text)

    if isWin then
        if rougelike.shouldShowUpgradeScreen(currentRun) then
            gamestate = GAMESTATES.upgrade
            rougelike.onLevelBeaten(currentRun)
            upgrade.init()
            return
        end

        rougelike.onLevelBeaten(currentRun)
    end

    gamestate = GAMESTATES.gameover

    gamestate.text = text
    gamestate.isWin = isWin
    gamestate.doUpdates = not isWin
    currentRun.endtime = os.time()
    leveldone.init(isWin, text)

end
GAMESTATES.default.doUpdates = true

GAMESTATES.paused = {handlequit = true, doUpdates = false}

GAMESTATES.paused.render = function(self)

    if camera then
        camera:draw(game.render)
    else
        game.render(0, 0, love.graphics.getWidth(), love.graphics.getHeight())
    end

    love.graphics.setColor(1, 1, 1, 1)
    -- renderFuelGauge
    local gaugeWidth = math.max(100, love.graphics.getWidth() / 10)
    love.graphics.setLineWidth(2)
    love.graphics.setColor(0.9, 0.9, 0.9, 1)
    love.graphics.rectangle("line", 20, 20, gaugeWidth, 24)
    love.graphics.setColor(21 / 255, 71 / 255, 52 / 255, 1)
    love.graphics.rectangle("fill", 22, 22,
                            (gaugeWidth - 2) * (player.fuel / player.maxFuel),
                            20)
    love.graphics.setColor(1, 1, 1, 1)
    love.graphics.print("Level: " .. currentRun.level, 0,
                        love.graphics.getHeight() - 30)

    love.graphics.print("FPS: " .. love.timer.getFPS(),
                        love.graphics.getWidth() - 100, 30)

    activeStage:render()
end

function GAMESTATES.paused.oninput(key)
    if key == "exit" then
        gamestate = GAMESTATES.default
        activeStage = nil
    else
        activeStage:input(key)
    end
end

GAMESTATES.paused.mousepressed = function(x, y, button)
    activeStage:mousepressed(x, y, button)
end
GAMESTATES.paused.mousemoved = function(x, y, dx, dy)
    activeStage:mousemoved(x, y, dx, dy)
end

GAMESTATES.gameover = {}
GAMESTATES.gameover.render = function(self)
    if camera then
        camera:draw(game.render)
    else
        game.render(0, 0, love.graphics.getWidth(), love.graphics.getHeight())
    end
    activeStage:render()

end

GAMESTATES.gameover.mousepressed = function(x, y, button)
    activeStage:mousepressed(x, y, button)
end
GAMESTATES.gameover.mousemoved = function(x, y, dx, dy)
    activeStage:mousemoved(x, y, dx, dy)
end

GAMESTATES.gameover.oninput = function(key)
    activeStage:input(key)
    if love.keyboard.isDown("lctrl") and key == "c" then
        local timeTaken = currentRun.endtime - currentRun.start
        local prettyTime = utils.prettyTime(timeTaken)
        local seed = tostring(currentRun.startseed)

        local newClipboardText = "You cleared " .. (currentRun.level - 1) ..
                                     " levels in " .. prettyTime .. " \nseed: '" ..
                                     seed .. "'"
        love.system.setClipboardText(newClipboardText)
    end
end

GAMESTATES.gameover.ongameover = function(...)
    -- ignore
end

GAMESTATES.upgrade = {}

GAMESTATES.upgrade.render = function(self)
    if camera then
        camera:draw(game.render)
    else
        game.render(0, 0, love.graphics.getWidth(), love.graphics.getHeight())
    end
    activeStage:render()
end

GAMESTATES.upgrade.mousepressed = function(x, y, button)
    activeStage:mousepressed(x, y, button)
end
GAMESTATES.upgrade.mousemoved = function(x, y, dx, dy)
    activeStage:mousemoved(x, y, dx, dy)
end

GAMESTATES.upgrade.oninput = function(key)

    local upgradeSucessful
    if key == "1" then
        upgradeSucessful = rougelike.upgrade(currentRun, "thrusters")
    elseif key == "2" then
        upgradeSucessful = rougelike.upgrade(currentRun, "fuel")
    elseif key == "3" then
        upgradeSucessful = rougelike.upgrade(currentRun, "reverse")
    end

    if upgradeSucessful then
        gamestate = GAMESTATES.default
        rougelike.generateLevel(currentRun)
        return
    end

    activeStage:input(key)

end

GAMESTATES.menu = {}
GAMESTATES.menu.render = function()
    
    if GAMESTATES.menu.background then love.graphics.draw(GAMESTATES.menu.background , 0, 0) end
    activeStage:render()
end

GAMESTATES.menu.oninput = function(key) activeStage:input(key) end

GAMESTATES.menu.mousepressed = function(x, y, button)
    activeStage:mousepressed(x, y, button)
end
GAMESTATES.menu.mousemoved = function(x, y, dx, dy)
    activeStage:mousemoved(x, y, dx, dy)
end

GAMESTATES.menu.inputseed = nil
GAMESTATES.menu.textinput = function(t) activeStage:textinput(t) end
GAMESTATES.credits = {handlequit = true}
GAMESTATES.credits.render = function(key) credits.render() end

GAMESTATES.credits.wheelmoved = function(x, y) credits.scroll(-y) end

GAMESTATES.credits.oninput = function(key)
    if key == "continue" or key == "exit" then
        gamestate = GAMESTATES.menu
        mainmenu.init()

        GAMESTATES.menu.inputseed = tostring(os.time())
    elseif key == "down" then
        credits.scroll(1)
    elseif key == "up" then
        credits.scroll(-1)
    end
end

function resetGameState()
    math.randomseed(currentRun.seed)
    game.reset()
    verlet.reset()
    gamestate = GAMESTATES.default
end
function love.load()
    sounsdstuff.init()

    love.window.maximize()
    All_Images = {
        smokeParticle = love.graphics.newImage('assets/smoke.png'),
        ship = love.graphics.newImage("assets/ship.png"),
        shipwreck1 = love.graphics.newImage("assets/ship_exploded_1.png"),
        shipwreck2 = love.graphics.newImage("assets/ship_exploded_2.png"),
        caveBackground = love.graphics.newImage("assets/caveBackground.png"),
        fuelCanister = love.graphics.newImage('assets/fuelCapsule.png'),
        explosion = love.graphics.newImage('assets/explosion.png'),
        explosionFlash = love.graphics.newImage('assets/explosionFlash.png')

    }

    loadspritesheet()

    defaultFont = love.graphics.newFont(20)
    love.graphics.setFont(defaultFont)

    ALLFONTS = {
        defaultPixel = love.graphics.newFont(
            "assets/fonts/PixelifySans-VariableFont_wght.ttf", 20),
        largePixel = love.graphics.newFont(
            "assets/fonts/PixelifySans-VariableFont_wght.ttf", 30),
        default = defaultFont
    }
    love.graphics.setFont(ALLFONTS.defaultPixel)
    
    SHADERS = {UI_HOVERING = love.graphics.newShader("shaders/green.glsl")}
    love.graphics.setDefaultFilter('nearest', 'nearest')

    gamestate = GAMESTATES.menu
    mainmenu.init()
    math.randomseed(os.time())
    GAMESTATES.menu.background = stars.generateCanvas(love.graphics.getWidth(), love.graphics.getHeight(), 50)



end

function startRougelikeGame()
    gamestate = GAMESTATES.default
    currentRun = rougelike.initRun(tonumber(GAMESTATES.menu.inputseed, 10))
    rougelike.generateLevel(currentRun)
end

function showCredits()
    activeStage = nil
    credits.init()
    gamestate = GAMESTATES.credits
end

function loadspritesheet()
    All_Images.rock = {}
    All_Images.rock.spritesheet = love.graphics
                                      .newImage('assets/rock/sheet.png')
    All_Images.rock.size = 150
    local size = All_Images.rock.size
    All_Images.rock.quads = {}
    local i = 0
    for y = 0, All_Images.rock.spritesheet:getHeight() - size, size do
        for x = 0, All_Images.rock.spritesheet:getWidth() - size, size do
            All_Images.rock.quads[i] = love.graphics.newQuad(x, y, size, size,
                                                             All_Images.rock
                                                                 .spritesheet)
            i = i + 1
        end

    end
end

function gameover(isWin, text) gamestate.ongameover(isWin, text) end

function love.draw()
    love.graphics.clear(0, 0, 0, 1)
    love.graphics.setColor(1, 1, 1, 1)
    if gamestate.render then gamestate:render() end
end

function love.update(dt)
    if gamestate.doUpdates then
        input.update(dt)
        game.update(dt)
        if level.update then level.update(dt) end
    end

end

function love.textinput(t)
    if gamestate and gamestate.textinput then gamestate.textinput(t) end
end

function love.joystickadded(joystick) p1joystick = joystick end

-- =====================
-- FPS independend physics

--[[
    Original Author: https://github.com/Leandros
    Updated Author: https://github.com/jakebesworth
        MIT License
        Copyright (c) 2018 Jake Besworth
    Original Gist: https://gist.github.com/Leandros/98624b9b9d9d26df18c4
    Love.run 11.X: https://love2d.org/wiki/love.run
    Original Article, 4th algorithm: https://gafferongames.com/post/fix_your_timestep/
    Forum Discussion: https://love2d.org/forums/viewtopic.php?f=3&t=85166&start=10
    Add this code to bottom of main.lua to override love.run() for Love2D 11.X
    Tickrate is how many frames your simulation happens per second (Timestep)
    Max Frame Skip is how many frames to allow skipped due to lag of simulation outpacing (on slow PCs) tickrate
---]]

local TICK_RATE = 1 / 100

-- How many Frames are allowed to be skipped at once due to lag (no "spiral of death")
local MAX_FRAME_SKIP = 25

-- No configurable framerate cap currently, either max frames CPU can handle (up to 1000), or vsync'd if conf.lua

function love.run()
    if love.load then love.load(love.arg.parseGameArguments(arg), arg) end

    -- We don't want the first frame's dt to include time taken by love.load.
    if love.timer then love.timer.step() end

    local lag = 0.0

    -- Main loop time.
    return function()
        -- Process events.
        if love.event then
            love.event.pump()
            for name, a, b, c, d, e, f in love.event.poll() do
                if name == "quit" then
                    if not love.quit or not love.quit() then
                        return a or 0
                    end
                end
                love.handlers[name](a, b, c, d, e, f)
            end
        end

        -- Cap number of Frames that can be skipped so lag doesn't accumulate
        if love.timer then
            lag = math.min(lag + love.timer.step(), TICK_RATE * MAX_FRAME_SKIP)
        end

        while lag >= TICK_RATE do
            if love.update then love.update(TICK_RATE) end
            lag = lag - TICK_RATE
        end

        if love.graphics and love.graphics.isActive() then
            love.graphics.origin()
            love.graphics.clear(love.graphics.getBackgroundColor())

            if love.draw then love.draw() end
            love.graphics.present()
        end

        -- Even though we limit tick rate and not frame rate, we might want to cap framerate at 1000 frame rate as mentioned https://love2d.org/forums/viewtopic.php?f=4&t=76998&p=198629&hilit=love.timer.sleep#p160881
        if love.timer then love.timer.sleep(0.001) end
    end
end
