-- Everything realated to the progression of a rougelike run (credits, challenge)
local levelgen = require("levelgen")
local stars = require("stars")
local progress = {}
progress.maxUpgrades = {fuel = 5, thrusters = 5, reverse = 3}

function progress.initRun(seed)
    local run = {
        credits = 0,
        level = 1,
        challenge = 1,
        startseed = seed or os.time(),
        start = os.time()
    }
    run.seed = run.startseed
    run.upgrades = {fuel = 0, thrusters = 0, reverse = 0}
    return run
end

function progress.upgrade(run, type)
    if not run.upgrades[type] then
        print("Error: unknown upgrade type")
        return false
    end

    if run.upgrades[type] < progress.maxUpgrades[type] then
        run.upgrades[type] = run.upgrades[type] + 1
        return true
    end

    return false

end

function progress.upgradesmaxed(run)
    for k, v in pairs(run.upgrades) do
        if progress.maxUpgrades[k] > v then return false end
    end
    return true

end

function progress.generateLevel(run)

    math.randomseed(run.seed)
    run.seed = math.random()

    resetGameState()
    verlet.width = 10000
    verlet.height = 10000
    local genLevel = levelgen.cellular.generate(progress.calculateSize(run), 4,
                                                1, 2)

    game.createLevelFromTileData(0, 300, genLevel.level)

    game.generateItems(0, 300, genLevel, progress.calculateItemCount(run), 150)

    local winArea = game.createArea((genLevel.furthestPoint.y) * 150 + 75,
                                    (genLevel.furthestPoint.x) * 150 + 300 + 75,
                                    50, function(area, obj)
        if obj == player then
            gameover(true, "You reached the target.")
            player:notmoving()
        end
    end)

    level = {name = "Level " .. run.level}

    local canvas = stars.generateCanvas(10000, 10000, 1000)
    game.background = canvas

    player = game.createRocket(350, 400, 210)
    player:applyUpgrades(run.upgrades)

    camera = gamera.new(0, 0, 10000, 10000)
end

function progress.calculateLength(run) return 10 * run.challenge end

function progress.calculateSize(run)
    return math.min(11 + (1 * run.challenge), 40)
end

function progress.calculateItemCount(run)
    return math.max(2, math.ceil(5 - (run.challenge / 2)))
end

function progress.onLevelBeaten(run)
    run.level = run.level + 1
    run.challenge = run.challenge + 0.5
end

function progress.shouldShowUpgradeScreen(run)
    return run.level >= 3 and run.level % 3 == 0 and
               not progress.upgradesmaxed(run)
end

return progress
