require("vector")
sat = {}

function sat.verts0(x, y, width, height)
    return {
        Vector:new(x, y), Vector:new(x + width, y),
        Vector:new(x + width, y + height), Vector:new(x, y + height)
    }
end

function sat.verts(x, y, width, height, rotation, center)
    return {
        Vector:new(x, y):rotateAround(rotation, center),
        Vector:new(x + width, y):rotateAround(rotation, center),
        Vector:new(x + width, y + height):rotateAround(rotation, center),
        Vector:new(x, y + height):rotateAround(rotation, center)
    }
end

function sat.axis(vert1, vert2) return (vert1 - vert2):normal():normalize() end

function sat.axislist(verts)
    local axislist = {}
    for i = 1, #verts - 1 do
        table.insert(axislist, sat.axis(verts[i], verts[i + 1]))
    end
    table.insert(axislist, sat.axis(verts[#verts], verts[1]))
    return axislist
end

function sat.projectc(circle, axis)
    local tmp = axis:scalarProduct(circle.pos)
    local pmin = tmp - circle.radius
    local pmax = tmp + circle.radius
    return {min = pmin, max = pmax}
end

function sat.project(verts, axis)
    local pmin = axis:scalarProduct(verts[1])
    local pmax = pmin
    for i = 2, #verts do
        local dot = axis:scalarProduct(verts[i])
        pmin = math.min(pmin, dot)
        pmax = math.max(pmax, dot)
    end

    return {min = pmin, max = pmax}
end

function sat.projectPoint(vert, axis) return axis:scalarProduct(vert) end

function sat._calcoverlap(minmax1, minmax2)
    return math.abs(math.max(minmax1.min, minmax2.min) -
                        math.min(minmax1.max, minmax2.max))
end

function sat.findClosestVert(verts, axis, other)
    local closestVert = nil
    local closestDistance = 1000000
    for i = 1, #verts do
        local dist = sat.projectPoint(verts[i], axis)
        if (dist < closestDistance) then
            closestVert = verts[i]
            closestDistance = dist;
        end
    end
    return closestVert
end

function sat._calcCollissionEdgePoints(axisIndex, verts)
    local to = axisIndex + 1
    if to == #verts + 1 then to = 1 end

    return {start = verts[axisIndex], endp = verts[to]}
end

function sat.circleAxixs(verts, circle)

    local mindist = 1000000
    local closest = nil
    local delta = nil
    for i, v in ipairs(verts) do
        delta = v - circle.pos
        local dist = delta:length()
        if dist < mindist then
            mindist = dist
            closest = delta

        end
    end

    return closest:normalize()
end

function sat.checkoverlapc(rect, circle)
    local verts = rect.verts
    local axisList = sat.axislist(verts)
    local circleAxixs = sat.circleAxixs(verts, circle)
    local smallestOverlap = nil
    local collissionAxis = nil

    local collissionVert = nil
    local colissionEdge = nil
    for i, axis in ipairs(axisList) do
        local minmax1 = sat.project(verts, axis)
        local minmax2 = sat.projectc(circle, axis)

        if minmax1.min > minmax2.max or minmax2.min > minmax1.max then
            return nil
        else
            local currentOverlap = sat._calcoverlap(minmax1, minmax2)
            if smallestOverlap == nil or currentOverlap < smallestOverlap then
                smallestOverlap = currentOverlap
                collissionAxis = axis
                collissionVert = circle.pos + (axis)
                collissionVert.obj = circle
                collissionEdge = sat._calcCollissionEdgePoints(i, verts)
                collissionEdge.obj = rect
            end
        end
    end

    local minmax1 = sat.project(verts, circleAxixs)
    local minmax2 = sat.projectc(circle, circleAxixs)

    if minmax1.min > minmax2.max or minmax2.min > minmax1.max then
        return nil
    else
        local currentOverlap = sat._calcoverlap(minmax1, minmax2)
        if smallestOverlap == nil or currentOverlap < smallestOverlap then
            smallestOverlap = currentOverlap
            collissionAxis = circleAxixs
            collissionVert = sat.findClosestVert(verts, circleAxixs, circle)
            collissionVert.obj = rect
            collissionEdge = {start = circle.pos, endp = circle.pos}
            collissionEdge.obj = circle
        end
    end
    return {
        axis = collissionAxis,
        overlap = smallestOverlap,
        vert = collissionVert,
        edge = collissionEdge
    }
end

function sat._checkoverlap(axisList, verts1, verts2, obj1, obj2, ret)
    for i, axis in ipairs(axisList) do
        local minmax1 = sat.project(verts1, axis)
        local minmax2 = sat.project(verts2, axis)

        if minmax1.min > minmax2.max or minmax2.min > minmax1.max then
            return nil
        else
            local currentOverlap = sat._calcoverlap(minmax1, minmax2)
            if ret.overlap == nil or currentOverlap < ret.overlap then
                ret.overlap = currentOverlap
                ret.axis = axis
                -- TODO sowohl vert als auch edge werden falsch berechnet unter gewissen bedingungen!!! ist schon calc overlap oder project( ) falsch???
                ret.vert = sat.findClosestVert(verts2, axis, obj1)
                ret.vert.obj = obj2
                ret.edge = sat._calcCollissionEdgePoints(i, verts1)
                ret.edge.obj = obj1
            end
        end
    end
    return ret
end

function sat.checkoverlap(obj1, obj2)
    local verts1 = obj1.verts
    local verts2 = obj2.verts
    local axisList1 = sat.axislist(verts1)
    local axisList2 = sat.axislist(verts2)
    local ret = {}

    ret = sat._checkoverlap(axisList1, verts1, verts2, obj1, obj2, ret)
    if ret == nil then return nil end

    ret = sat._checkoverlap(axisList2, verts2, verts1, obj2, obj1, ret)
    return ret

end
