local ui = {}

ui.states = {}
ui.states.normal = {}

function ui.states.normal.mouseentered(element)
    element.state = ui.states.hovering
    element.shader = SHADERS.UI_HOVERING
end
function ui.states.normal.mouseexited(element)
    -- nix
end

function ui.states.normal.select(element)
    element.state = ui.states.hovering
    element.shader = SHADERS.UI_HOVERING
end

function ui.states.normal.deselect(element)
    -- nix
end

function ui.states.normal.clicked(element)
    -- nix
end

ui.states.hovering = {}
function ui.states.hovering.mouseexited(element)
    element.state = ui.states.normal
    element.shader = nil
end

function ui.states.hovering.mouseentered(element)
    -- nix
end
function ui.states.hovering.clicked(element, button)
    if element.onclick then element.onclick(button) end
end

function ui.states.hovering.select(element)
    -- nix
end

function ui.states.hovering.deselect(element)
    element.state = ui.states.normal
    element.shader = nil
end

function ui.newLabel(text, x, y)
    local element = {x = x or 0, y = y or 0, text = text, state = ui.states.normal}
    local _, lines = text:gsub("\n", "")
    lines = lines + 1
    element.height = love.graphics.getFont():getHeight() * lines
    element.width = love.graphics.getFont():getWidth(element.text)

    function element:render(offsetX, offsetY)
        love.graphics.setColor(1, 1, 1, 1)
        local textwidth = love.graphics.getFont():getWidth(self.text)
        local textHeight = love.graphics.getFont():getHeight() * lines
        love.graphics.print(self.text,
                            self.x + offsetX + self.width / 2 - textwidth / 2,
                            self.y + offsetY + self.height / 2 - textHeight / 2)

    end

    function element:mouseEntered() end

    function element:clicked(button) end

    function element:mouseExited() end

    function element:select() end

    function element:deselect() end

    return element
end

function ui.newTextField(width, initialText, textColor, x, y)
    local element = {
        x = x or 0,
        y = y or 0,
        width = width,
        text = initialText,
        state = ui.states.normal,
        textColor = textColor or {r = 1, g = 1, b = 1, a = 1},
        focusable = true
    }

    element.height = love.graphics.getFont():getHeight() + 8
    function element:render(offsetX, offsetY)

        love.graphics.setColor(self.textColor.r, self.textColor.g,
                               self.textColor.b, self.textColor.a)

        local textwidth = love.graphics.getFont():getWidth(self.text)
        local textHeight = love.graphics.getFont():getHeight()
        love.graphics.setShader(element.shader)
        love.graphics.rectangle("line", self.x + offsetX, self.y + offsetY,
                                self.width, self.height)
        love.graphics.setShader()

        love.graphics.print(self.text, self.x + offsetX + 5, self.y + offsetY +
                                (self.height) / 2 - textHeight / 2)

    end

    function element:mouseEntered() end

    function element:clicked(button) self.state.mouseentered(self) end

    function element:mouseExited() end

    function element:select() self.state.select(self) end

    function element:deselect() self.state.deselect(self) end

    function element:input(key)
        if key == "delete" then
            if string.len(self.text) > 0 then
                self.text = self.text:sub(1, -2)
            end
        elseif love.keyboard.isDown("lctrl") and key == "v" then
            local clipboardtext = love.system.getClipboardText()
            if tonumber(clipboardtext, 10) ~= nil then
                self.text = clipboardtext:sub(1, 10)
            end
        elseif love.keyboard.isDown("lctrl") and key == "c" then
            love.system.setClipboardText(self.text)

        end
    end

    return element
end

function ui.newButton(text, x, y, width, height)
    local element = {
        x = x or 0,
        y = y or 0,
        width = width or 0,
        height = height or 0,
        text = text,
        state = ui.states.normal,
        focusable = true
    }

    local _, lines = text:gsub("\n", "")
    lines = lines + 1
    element.height = math.max(love.graphics.getFont():getHeight() * lines, element.height)
    element.width = math.max(
                        love.graphics.getFont():getWidth(element.text) + 10,
                        element.width)


    function element:render(offsetX, offsetY)
        love.graphics.setColor(1, 1, 1, 1)
        local textwidth = love.graphics.getFont():getWidth(self.text)
        local textHeight = love.graphics.getFont():getHeight()
        love.graphics.setShader(element.shader)
        love.graphics.rectangle("line", self.x + offsetX, self.y + offsetY,
                                self.width, self.height)
        love.graphics.setShader()

        love.graphics.print(self.text,
                            self.x + offsetX + self.width / 2 - textwidth / 2,
                            self.y + offsetY + (self.height) / 2 - textHeight /
                                2)

    end

    function element:mouseEntered() self.state.mouseentered(self) end

    function element:clicked(button) self.state.clicked(self, button) end

    function element:mouseExited() self.state.mouseexited(self) end

    function element:select() self.state.select(self) end

    function element:deselect() self.state.deselect(self) end

    return element
end

function ui.mouseInElement(x, y, element)
    return x > element.x and y > element.y and x < element.x + element.width and
               y < element.y + element.height
end

function ui.linearVerticalLayoutManager(container)

    local currentX = container.margin
    local currentY = container.margin

    local maxWidth = container.width

    for i, child in ipairs(container.children) do
        child.x = currentX
        child.y = currentY
        child.width = math.max(container.width - container.margin * 2,
                               child.width or 0)
        maxWidth = math.max(maxWidth, child.width + container.margin * 2)

        if child.children then
            local containerX, containerY = child:layout()
            currentY = currentY + containerY
        end

        currentX, currentY = child.x,
                             currentY + child.height + container.padding
    end

    return maxWidth - container.padding, currentY + container.margin - container.padding

end


function ui.linearHorizontalLayoutManager(container)

    local currentX = container.margin
    local currentY = container.margin
    local maxHeight = container.height

    for i, child in ipairs(container.children) do
        child.x = currentX
        child.y = currentY
        child.height = math.max(container.height - container.margin * 2,
                                child.height or 0)
        maxHeight = math.max(maxHeight, child.height + container.margin * 2)

        if child.children then
            local containerX, containerY = child:layout()
            currentX = currentX + containerX
        end

        currentX, currentY = child.x + child.width + container.padding, currentY
    end

    return currentX + container.margin - container.padding, maxHeight - container.padding

end

function ui.newPanel(x, y, maxX, maxY, padding, margin)
    local container = {
        x = x,
        y = y,
        maxX = maxX,
        maxY = maxY,
        currentX,
        currentY,
        padding = padding or 5,
        margin = margin or 5
    }
    container.height = maxY
    container.width = maxX
    container.layoutManager = ui.linearVerticalLayoutManager
    container.children = {}
    function container:addCild(child)
        table.insert(self.children, child)

        return #self.children
    end

    function container:render(offsetX, offsetY)
        if self.bgcolor then
            love.graphics.setColor(self.bgcolor.r, self.bgcolor.g,
                                   self.bgcolor.b, self.bgcolor.a)
            love.graphics.rectangle("fill", self.x + offsetX, self.y + offsetY,
                                    (self.width or self.maxX) + self.margin,
                                    (self.height or self.maxY) + self.margin)
        end

        for i, v in ipairs(self.children) do
            v:render(offsetX + self.x, offsetY + self.y)
        end
    end

    function container:mouseEntered(x, y)
        for i, v in ipairs(self.children) do
            if ui.mouseInElement(x - self.x, y - self.y, v) then
                v:mouseEntered(x - self.x, y - self.y)
            else
                v:mouseExited()
            end

        end
    end

    function container:mouseExited()
        for i, v in ipairs(self.children) do v:mouseExited() end
    end
    function container:clicked() end

    function container:layout() return self.layoutManager(self) end



function container:pack()
    for i,v in ipairs(self.children) do
        if v.children then
            v:pack()
        end
    end
    self.width, self.height = self:layout()

end

    return container
end

function ui.newStage()
    local stage = {elements = {}}

    function stage:addElement(element)
        table.insert(self.elements, element)
        if element.children then
            for k, v in pairs(element.children) do self:addElement(v) end
        end

        local lastElement = nil
        local firstElement = nil
        if element.focusable then
            for i = 1, #self.elements do
                if self.elements[i].focusable then
                    if firstElement == nil then
                        firstElement = self.elements[i]
                    end
                    if lastElement == nil then
                        lastElement = self.elements[i]
                    end
                    lastElement.next = self.elements[i]
                    self.elements[i].prev = lastElement
                    lastElement = self.elements[i]
                end

            end
            lastElement.next = firstElement
            firstElement.prev = lastElement

            self.firstSelectable = firstElement
            self.lastSelectable = lastElement
        end

    end

    function stage:addRootElement(element)
        self.root = element
        self:addElement(element)

    end

    function stage:mousepressed(x, y, button)
        for k, v in pairs(self.elements) do v:clicked(button) end
    end
    function stage:mousemoved(x, y, dx, dy)

        if self.root then
            if ui.mouseInElement(x, y, self.root) then
                self.root:mouseEntered(x, y)
            else
                self.root:mouseExited()
            end
        end
    end

    function stage:getSelectedIndex()
        for i = 1, #self.elements do
            if self.elements[i].state == ui.states.hovering then
                return i
            end
        end
        return nil
    end

    function stage:input(key)
        if #self.elements == 0 then
            -- NO elements, nothing to do
            return
        end
        if key == "down" or key == "right" then
            local selectedIndex = self:getSelectedIndex()
            if not selectedIndex then
                if self.firstSelectable then
                    self.firstSelectable:select()
                end

            else
                self.elements[selectedIndex]:deselect()
                self.elements[selectedIndex].next:select()
            end
        elseif key == "up" or key == "left" then
            local selectedIndex = self:getSelectedIndex()
            if not selectedIndex then
                if self.lastSelectable then
                    self.lastSelectable:select()
                end
            else
                self.elements[selectedIndex]:deselect()
                self.elements[selectedIndex].prev:select()
            end
        elseif key == "continue" or key == "start" then
            local selectedIndex = self:getSelectedIndex()
            if selectedIndex then
                -- Act like it was a primary button click
                self.elements[selectedIndex]:clicked(1)
            end

        else
            local selectedIndex = self:getSelectedIndex()
            if selectedIndex and self.elements[selectedIndex].input then
                self.elements[selectedIndex]:input(key)
            end
        end

    end
    function stage:render()
        local fontbefore = love.graphics.getFont()

        love.graphics.setFont(ALLFONTS.defaultPixel)
        love.graphics.setColor(1, 1, 1, 1)

        if self.root then self.root:render(0, 0) end

        love.graphics.setFont(fontbefore)

    end

    function stage:textinput(t)
        local selectedIndex = self:getSelectedIndex()
        if selectedIndex and self.elements[selectedIndex].textinput then
            self.elements[selectedIndex]:textinput(t)
        end
    end

    return stage
end
return ui

