-- Mutable vector
Vector = {}

function Vector:new(x, y)
    newCircle = {x = x or 0, y = y or 0}
    self.__index = self
    return setmetatable(newCircle, self)
end
Vector.NORMAL_VECTOR = Vector:new(1)

function Vector:add(other)
    self.x = self.x + other.x
    self.y = self.y + other.y
    return self
end

function Vector:set(other)
    self.x = other.x
    self.y = other.y
end

function Vector:setxy(newX, newY)
    self.x = newX
    self.y = newY
end

function Vector.__add(v1, v2)
    local result = v1:clone()
    return result:add(v2)
end

function Vector.__mul(v, scalar) return v:clone():mult(scalar) end

function Vector.__sub(v1, v2)
    local result = v1:clone()
    return result:minus(v2)
end

function Vector:minus(other)
    self.x = self.x - other.x
    self.y = self.y - other.y
    return self
end

function Vector:mult(scalar)
    self.x = self.x * scalar
    self.y = self.y * scalar
    return self
end

function Vector:scalarProduct(other) return self.x * other.x + self.y * other.y end

function Vector:angleTo(other)
    -- in radians
    local angle = math.acos(self:scalarProduct(other) /
                                (self:length() * other:length()))

    -- degree wäre / PI  * 180
    return angle
end

function Vector:distance(other)
    return math.sqrt((self.x - other.x) ^ 2 + (self.y - other.y) ^ 2)
end

function Vector:distanceSquared(other)
    return (self.x - other.x) ^ 2 + (self.y - other.y) ^ 2
end
function Vector:clone() return Vector:new(self.x, self.y) end

function Vector:length() return math.sqrt(self.x ^ 2 + self.y ^ 2) end

function Vector:normalize()
    local length = self:length()
    self.x = self.x * (1 / length)
    self.y = self.y * (1 / length)
    return self
end

function Vector:normal() return Vector:new(-self.y, self.x) end

function Vector:rotate(angle)
    -- angle in radians
    -- positive = counterclockwise, negative = clockwise

    local newX = self.x * math.cos(angle) - self.y * math.sin(angle)
    local newY = self.x * math.sin(angle) + self.y * math.cos(angle)
    self.x = newX
    self.y = newY
    return self
end

function Vector:rotateAround(angle, point)
    self:minus(point)
    self:rotate(angle)
    self:add(point)
    return self
end

function Vector:toString() return "(" .. self.x .. "," .. self.y .. ")" end
